#####################################################################
# Copyright (C) 2000 Artur Trzewik
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
######################################################################
# Header
# CVS - $Id: sqlschema_ext.tcl,v 1.8 2000/09/28 18:24:10 artur Exp $
#
# Information:
# Modul (benutzt von Schema Editor)
# Diese Klassen lesen das DB Schema mit Hilfe von Sql aus.
# Das Auslese-Proze geschiet mit Hilfe von Beusucher Muster.
# Besucher-Objekt soll die Schnittstelle besitzen:
# newtable {tablename}
# newattribute parlist->{{name ??} {not_null ?} {type ?} ...}
#
Class MySqlSchemaExtractor -superclass MySql -parameter {
    {dbspecific {name not_null primary_key default type auto_increment}}
}
MySqlSchemaExtractor instproc extract_schema {visitor} {
    set qhandle [[self] query "show tables"]
    set tables ""
    while {[set table [$qhandle fetch]]!=""} {
	lappend tables $table
    }
    $qhandle destroy
    foreach table $tables {
	$visitor newtable $table
	[self] extract_table $visitor $table
    }
}
MySqlSchemaExtractor instproc extract_table {visitor table} {
    set qhandle [[self] query "describe $table"]
    while {[set attr [$qhandle fetch]]!=""} {
	set alist {}
	keylset alist name [lindex $attr 0]
	# Der Typ wird in der Form angegeben typ(length) z.B varchar(20)
	# Hier wird er in zwei Attribute gespalten
	if {[regexp {([a-z]+)\((.+)\)} [lindex $attr 1] _ typ length]} {
	    keylset alist type $typ
	    if {$typ=="enum" || $typ=="set"} {
		set vl {}
		foreach it [split $length ,] {
		    lappend vl [string trim $it '\"]
		}
		keylset alist valuelist $vl
	    } else {
		keylset alist length $length
	    }
	} else {
	    keylset alist type [lindex $attr 1]
	}
	if {[lindex $attr 2]=="YES"} {
	    keylset alist null 1
	}
	if {[lindex $attr 3]=="PRI"} {
	    keylset alist primary_key 1
	}
	if {[lindex $attr 4]!=""} {
	    keylset alist default [lindex $attr 4]
	}
	if {[lindex $attr 5]=="auto_increment"} {
	    keylset alist auto_increment 1
	}
	$visitor newattribute $alist
    }
    $qhandle destroy
}
# Object DummyVisitor
# DummyVisitor proc newtable table {
#puts "Table $table"
# }
# DummyVisitor proc newattribute attr {
#puts "       $attr"
# }
Class PgSqlSchemaExtractor -superclass PgSql -parameter {
    {dbspecific {name not_null primary_key default type auto_increment}}
}
PgSqlSchemaExtractor instproc extract_schema {visitor} {
    # Die pga_* Tabellen werden nicht bercksichtigt
    # Abgeschaut von pgaccess Datei databases.tcl Procedur  CurrentDB 
    set qhandle [[self] query {select relname from pg_class where (relname !~ '^pg_') and (relkind='r') and  (relname !~ '^pga_')}]
    set tables ""
    while {[set table [$qhandle fetch]]!=""} {
	lappend tables $table
    }
    $qhandle destroy
    foreach table $tables {
	$visitor newtable $table
	[self] extract_table $visitor $table
    }
}
PgSqlSchemaExtractor instproc extract_table {visitor table} {
    #puts "extact Table";
    set qhandle [[self] query "select attname,typname,attnotnull,atttypmod,attlen from pg_class,pg_attribute,pg_type where pg_class.relname='$table' and pg_class.oid=pg_attribute.attrelid and pg_attribute.atttypid=pg_type.oid and attnum>0 order by attnum"]
    while {[set attr [$qhandle fetch]]!=""} {
	#puts "extact Table --";
	set alist {}
	keylset alist name [lindex $attr 0]
	keylset alist type [lindex $attr 1]
	if {[lindex $attr 2]=="f"} {
	    keylset alist null 1
	}
	if {[lindex $attr 4]=="-1" && [lindex $attr 3]!="-1" } {
	    keylset  alist tlenght [expr [lindex $attr 3]-4]
	}
	$visitor newattribute $alist
    }
    $qhandle destroy
}
